var http = require("http");
var path = require("path");
var responseBody = require("./util").responseBody;

var buster = require("buster-core");
var bPromise = require("buster-promise");
var bSession = require("./session");

function requestSessionCreation(client, config, callback) {
    var req = http.request(client.httpOptions("POST", "sessions"), function (res) {
        bSession.createFromHttpResponse(res, {
            host: client.host,
            port: client.port
        }).then(function (session) {
            callback(null, session);
        });
    });

    var resources = config.resources || [];

    for (var path in resources) {
        delete resources[path].cacheable;
    }

    req.on("error", callback);
    req.write(JSON.stringify(config));
    req.end();
}

function isCacheable(resource) {
    return resource.cacheable == null || !!resource.cacheable;
}

function isCached(path, resources, cache) {
    return cache[path] && cache[path].indexOf(resources[path].etag) >= 0;
}

module.exports = {
    cacheResources: true,

    create: function (port, host, path) {
        var client = Object.create(this);
        client.port = port || 1111;
        client.host = host || "localhost";
        client.path = path || "/";

        return client;
    },

    getCachedResources: function () {
        var promise = bPromise.create();
        var self = this;

        if (this.cachedResources) {
            promise.resolve(this.cachedResources);
        } else {
            var opt = this.httpOptions("GET", "resources");

            var req = http.request(opt, function (res) {
                if (res.statusCode != 200) {
                    return promise.reject(new Error(
                        "Server responded to cache lookup with " +
                            res.statusCode + ", expected 200\n" +
                            "Is this server (" + opt.host + ":" + opt.port +
                            ") a buster-server instance?"));
                }

                responseBody(res, function (body) {
                    try {
                        self.cachedResources = body && JSON.parse(body) || {};
                        promise.resolve(self.cachedResources);
                    } catch (e) {
                        promise.reject(new Error(
                            "Server responded to cache lookup with " +
                                "invalid response data. Is this server (" +
                                opt.host + ":" + opt.port +
                                ") a buster-server instance?\n" +
                                "Original error was " + e.message));
                    }
                });
            }).on("error", function (err) {
                promise.reject(err);
            }).end();
        }

        return promise;
    },

    useCachedResources: function (config) {
        var promise = bPromise.create();
        var resources = config.resources;

        this.getCachedResources().then(function (cached) {
            for (var path in resources) {
                if (isCacheable(resources[path]) &&
                    isCached(path, resources, cached)) {
                    delete resources[path].content;
                }
            }

            promise.resolve(config);
        }, function (err) {
            promise.reject(err);
        });

        return promise;
    },

    createSession: function (sessionConfig) {
        var promise = bPromise.create();
        var self = this;

        function configLoaded(config) {
            requestSessionCreation(self, config, function (err, session) {
                if (err) return promise.reject(err);
                promise.resolve(session);
            });
        }

        sessionConfig.configure().then(function (config) {
            if (self.cacheResources) {
                self.useCachedResources(config).then(configLoaded, function (err) {
                    promise.reject(err);
                });
            } else {
                configLoaded(config);
            }
        }, function (err) {
            err.message = "Bad session config: " + err.message;
            promise.reject(err);
        });

        return promise;
    },

    httpOptions: function (method, rpath) {
        return {
            method: method,
            host: this.host,
            port: this.port,
            path: path.join(this.path, rpath),
            headers: {
                "Content-Type": "application/json",
                "Accept": "application/json"
            }
        };
    }
};
