var buster = require("buster");
var assert = buster.assert;
var refute = buster.refute;
var createServerFor = require("../test-helper").createServerFor;
var bSession = require("../../lib/buster-client/session");

buster.testCase("buster.session", {
    "createFromHttpResponse": {
        setUp: function () {
            this.response = buster.extend(buster.create(buster.eventEmitter), {
                headers: {
                    connection: "close",
                    "transfer-encoding": "chunked",
                    location: "/sessions/1"
                },
                statusCode: 201,
                setEncoding: this.spy()
            });

            this.respond = function (response) {
                response = response || JSON.stringify({
                    rootPath: "/sessions/1",
                    multicastUrl: "/sessions/1/messages",
                    multicastClientId: 1
                });

                process.nextTick(function () {
                    this.response.emit("data", response);
                    this.response.emit("end");
                }.bind(this));

                return this.response;
            };

            this.stub(buster.multicastClient, "listen");
        },

        "should resolve promise with session object": function (done) {
            var instance = { connect: this.spy() };
            this.stub(bSession, "create").returns(instance);

            bSession.createFromHttpResponse(this.response, {
                host: "localhost",
                port: 1111
            }).then(function (session) {
                assert.calledWith(response.setEncoding, "utf8");
                assert.calledWith(bSession.create, {
                    rootPath: "/sessions/1",
                    multicastUrl: "/sessions/1/messages",
                    multicastClientId: 1,
                    host: "localhost",
                    port: 1111
                });

                assert.same(instance, session);
                done();
            });

            var response = this.respond();
        },

        "should reject promise with JSON parse error": function (done) {
            this.spy(bSession, "create");

            bSession.createFromHttpResponse(this.response).then(
                function () {},
                function (err, session) {
                    assert.isUndefined(session);
                    refute.isNull(err);
                    assert.notCalled(bSession.create);
                    done();
                });

            var response = this.respond("{");
        },

        "should connect multicaster": function (done) {
            bSession.createFromHttpResponse(this.response).then(function () {
                assert.calledOnce(buster.multicastClient.listen);
                done();
            });

            var response = this.respond();
        },

        "should not connect multicaster when no multicast path": function (done) {
            bSession.createFromHttpResponse(this.response).then(function (s) {
                assert.isUndefined(s.multicaster);
                assert.notCalled(buster.multicastClient.listen);
                done();
            });

            var response = this.respond(JSON.stringify({
                rootPath: "/sessions/1"
            }));
        }
    },

    "create": {
        "should set rootPath, host and port": function () {
            var session = bSession.create({
                rootPath: "/sessions/0",
                host: "buster.local",
                port: 1110
            });

            buster.assert.match(session, {
                rootPath: "/sessions/0",
                host: "buster.local",
                port: 1110
            });
        },

        "should use default host": function () {
            var session = bSession.create({
                rootPath: "/sessions/0",
                port: 1110
            });

            buster.assert.equals(session.host, "localhost");
        },

        "should use default port": function () {
            var session = bSession.create({ rootPath: "/sessions/0" });

            buster.assert.equals(session.port, 1111);
        }
    },

    "close": {
        setUp: function (done) {
            this.requests = createServerFor(this, done);
        },

        tearDown: function () {
            this.server.close();
        },

        "should make DELETE request to session": function (done) {
            var session = bSession.create({
                rootPath: "/sessions/1",
                host: "localhost",
                port: 7777
            });

            session.close().then(function () {
                assert.equals(this.requests.length, 1);
                assert.match(this.requests[0], {
                    method: "DELETE",
                    url: "/sessions/1"
                });
                
                done();
            }.bind(this));
        },

        "should use session properties in DELETE request": function (done) {
            var session = bSession.create({
                rootPath: "/sessions/2",
                port: 7777
            });

            session.close().then(function () {
                assert.match(this.requests[0], { url: "/sessions/2" });
                done();
            }.bind(this));
        },

        "should reject promise with error on bad connection": function (done) {
            var session = bSession.create({
                rootPath: "/sessions/2",
                host: "bogus"
            });

            session.close().then(function () {}, function (err) {
                refute.isNull(err);
                assert.match(err.message, /(Domain name not found)|(Could not contact DNS servers)/);
                done();
            }.bind(this));
        },

        "should disconnect multicaster": function (done) {
            var session = bSession.create({
                rootPath: "/sessions/2",
                port: 7777
            });

            var disconnect = this.spy();
            session.multicaster = { disconnect: disconnect };

            session.close().then(function () {
                assert.calledOnce(disconnect);
                done();
            }.bind(this));
        },

        "should kill session when disconnecting multicaster is done": function (done) {
            var session = bSession.create({
                rootPath: "/sessions/2",
                port: 7777
            });

            var http = require("http");
            this.spy(http, "request");
            var promise = buster.promise.create();
            var disconnect = this.stub().returns(promise);
            session.multicaster = { disconnect: disconnect };

            session.close();

            assert.calledOnce(disconnect);

            process.nextTick(function () {
                assert.notCalled(http.request);
                promise.resolve();

                process.nextTick(function () {
                    assert.calledOnce(http.request);
                    done();
                });
            });
        }
    },

    "multicasting": {
        setUp: function (done) {
            var mc = this.multicaster = Object.create(buster.multicastMiddleware);
            mc.contextPath = "/sessions/1/messages";

            this.server = require("http").createServer(function (req, res) {
                mc.respond(req, res);
            });

            this.server.listen(7777, done);
        },

        tearDown: function () {
            this.session.close();
            this.server.close();
        },

        "should send message with multicast client": function (done) {
            var multicaster = this.multicaster.createClient();

            this.session = bSession.create({
                port: 7777,
                multicastUrl: multicaster.url,
                multicastClientId: multicaster.clientId
            });

            this.session.connect();

            this.session.on("msg", function (msg) {
                assert.equals(msg.data, 42);
                done();
            });

            this.session.emit("msg", 42);
        },

        "on should fail if session has no multicasting": function () {
            this.session = bSession.create({
                port: 7777
            });

            try {
                this.session.on("msg", function () {});
                throw new Error("Expected to fail");
            } catch (e) {
                assert.match(e.message, "Session has no multicaster");
            }
        },

        "emit should fail if session has no multicasting": function () {
            this.session = bSession.create({
                port: 7777
            });

            try {
                this.session.emit("msg");
                throw new Error("Expected to fail");
            } catch (e) {
                assert.match(e.message, "Session has no multicaster");
            }
        },

        "should subscribe to any event": function (done) {
            var multicaster = this.multicaster.createClient();

            this.session = bSession.create({
                port: 7777,
                multicastUrl: multicaster.url,
                multicastClientId: multicaster.clientId
            });

            this.session.connect();

            this.session.onMessage(function (msg) {
                assert.equals(msg.data, 42);
                assert.equals(msg.topic, "msg");
                done();
            });

            this.session.emit("msg", 42);
        },

        "onMessage should fail if session has no multicasting": function () {
            this.session = bSession.create({
                port: 7777
            });

            try {
                this.session.onMessage(function () {});
                throw new Error("Expected to fail");
            } catch (e) {
                assert.match(e.message, "Session has no multicaster");
            }
        }
    },

    "getting clients": {
        setUp: function () {
            this.session = bSession.create({ port: 7777 });
            this.response = { on: this.stub() };
            var end = this.spy();

            this.request = this.stub(require("http"), "request").returns({
                end: end
            });

            this.request.end = end;
        },

        "should return promise": function () {
            var promise = this.session.getClients();

            assert.isObject(promise);
            assert.isFunction(promise.then);
        },

        "should GET /sessions/messaging/clients": function () {
            var promise = this.session.getClients();

            assert.calledOnce(this.request);
            assert.equals(this.request.args[0][0], {
                path: "/sessions/messaging/clients",
                host: "localhost",
                port: 7777,
                agent: false
            });
        },

        "should resolve promise with server response": function (done) {
            var clients = [
                { id: 1, url: "/msg/1", agent: { browser: "Firefox" } },
                { id: 2, url: "/msg/2", agent: { browser: "Chrome" } }];

            this.response.on.yields(JSON.stringify(clients));
            this.request.yields(this.response);

            this.session.getClients().then(function (clients) {
                assert.equals(clients, [
                    { id: 1, url: "/msg/1", agent: { browser: "Firefox" } },
                    { id: 2, url: "/msg/2", agent: { browser: "Chrome" } }]);
                done();
            });
        }
    }
});
